/*
	Balloon.h 
	Routines for drawing balloons
*/

class Balloon
{
public:
	Balloon();
	
	Vector pos, vel;
	String string;
	Balloon *next;
};

Balloon::Balloon()
{
	next = NULL;
}

Vector Hermite(float s) //Draws a hermite spline, based on two points and the tangent line at each point
{
	Vector vec;
	float s2 = s*s;
	float s3 = s*s*s;

	vec.x = ((2*s3 - 3*s2 + 1)*p1.x) + ((-2*s3 + 3*s2)*p2.x) + ((s3 - 2*s2 + s)*t1.x) + ((s3 - s2)*t2.x);
	vec.y = ((2*s3 - 3*s2 + 1)*p1.y) + ((-2*s3 + 3*s2)*p2.y) + ((s3 - 2*s2 + s)*t1.y) + ((s3 - s2)*t2.y);
	vec.z = ((2*s3 - 3*s2 + 1)*p1.z) + ((-2*s3 + 3*s2)*p2.z) + ((s3 - 2*s2 + s)*t1.z) + ((s3 - s2)*t2.z);

	return vec;
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////
class Spline  //This is just a vector (array) class here to simplify the use of 2D arrays
{
public:
	Spline(int divs);
	Spline();

	Vector &operator [] (int k);
	
	Vector *points;
};

Spline::Spline() {}

Spline::Spline(int divs)
{
	points = new Vector[divs];
}

Vector &Spline::operator [] (int k)
{
	return points[k];
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void draw_balloon(int detail) //This code is an unholy mess, but it only needs to be run a few times 
{
	Spline point(detail+1), *balloon, *normals;
	balloon = new Spline[detail];
	normals = new Spline[detail];

	for (int i = 0; i < detail; i++) //Initialize the two arrays
	{
		balloon[i] = Spline(detail);
		normals[i] = Spline(detail);
	}
	
	float t;
	for (i = 0; i <= detail; i++) //Create the balloon "curve" that will later be rotated
	{
		t = i / float(detail);
		point[i] = Hermite(t);
		if ((i!=0) && (i!=detail)) balloon[0][i-1] = point[i];
	}
	
	
	Vector bottom = point[detail-1], top = point[0];
	rotate = rotateY(360/(float)detail); //Rotate the curve around to create all the balloon's vertices
	for (i = 0; i < detail-1; i++)
		for (int j = 1; j < detail; j++)
		{
			balloon[j][i] = rotate * balloon[j-1][i];
		}
	
	//The following bit calculates all the normals for the vertices. This code is evil, but it works
	//and I'm unwilling to screw it up by trying to optimize it. 
	Vector dir[4];
	Vector norm[4];
	int u, v, dx=0, dy=0;
	for (i = 0; i < detail; i++)  //Go through each vertex
		for (int j = 0; j < detail-1; j++)
		{
			for (int k = 0; k < 4; k++) //Then, get the normals of the four adjacent faces
			{
				k < 2 ? (k == 0 ? dx=1 : dy=1) : (k == 2 ? dx=-1 : dy=-1);

				u = (i+dx > detail-1) ? 1 : (i+dx < 0) ? detail-1 : i+dx;			
				v = (j+dy > detail-2) ? -1 : (j+dy < 1) ? -2 : j+dy;

				if (v==-1) dir[k] = balloon[i][j]-bottom;
				else if (v==-2) dir[k] = balloon[i][j]-top;
				else dir[k] = balloon[i][j]-balloon[u][v];

				dx = dy = 0;
			}

			norm[0] = cross(dir[1], dir[0]);
			norm[1] = cross(dir[2], dir[1]);
			norm[2] = cross(dir[3], dir[2]);
			norm[3] = cross(dir[0], dir[3]);

			for (int x = 0; x < 4; x++)  //Then, average all the normals together to get the vertice's normal
				normals[i][j] += norm[x];

			normals[i][j] = normals[i][j].normal(); //hehe, normal...
		}

	glBegin(GL_TRIANGLE_FAN); //Draw top-most section 
		glNormal3f(0,1,0);
		glVertex3f(point[0].x,point[0].y,point[0].z); 
		for (int k = 0; k < detail; k++)
		{
			glNormal3f(normals[k][1].x,normals[k][1].y,normals[k][1].z);
			glVertex3f(balloon[k][1].x,balloon[k][1].y,balloon[k][1].z);
		}
		glNormal3f(normals[0][1].x,normals[0][1].y,normals[0][1].z);
		glVertex3f(balloon[0][1].x,balloon[0][1].y,balloon[0][1].z);
	glEnd();
	
	for (int j = 0; j < detail; j++)
	{
		glBegin(GL_TRIANGLE_STRIP); //Draw main portion of balloon
		for (int k = 1; k < detail-1; k++)
		{
			glNormal3f(normals[j][k].x,normals[j][k].y,normals[j][k].z);
			glVertex3f(balloon[j][k].x,balloon[j][k].y,balloon[j][k].z);
			glNormal3f(normals[(j+1)%detail][k].x,normals[(j+1)%detail][k].y,normals[(j+1)%detail][k].z);							
			glVertex3f(balloon[(j+1)%detail][k].x,balloon[(j+1)%detail][k].y,balloon[(j+1)%detail][k].z);							
		}
		glEnd();
	}
	
	if (normal) //This is a debug routine I'm keeping in just in case someone it's needed at some point
	{
		glBegin(GL_LINES);
		glVertex3f(top.x, top.y, top.z);
		glVertex3f(top.x, top.y+1, top.z);
		for (j = 0; j < detail; j++)
		{
			for (int k = 1; k < detail-1; k++)
			{
				glVertex3f(balloon[j][k].x,balloon[j][k].y,balloon[j][k].z);
				glVertex3f(balloon[j][k].x+(normals[j][k].x)*.3,balloon[j][k].y+(normals[j][k].y)*.3,balloon[j][k].z+(normals[j][k].z)*.3);
			}
		}
		glEnd();
	}

	glBegin(GL_TRIANGLE_FAN);
		glNormal3f(0,-1,0);
		glVertex3f(point[detail].x,point[detail].y,point[detail].z);  //Draw bottom-most section
		for (k = 0; k < detail; k++)
		{
			glNormal3f(normals[k][detail-2].x,normals[k][detail-2].y,normals[k][detail-2].z);
			glVertex3f(balloon[k][detail-2].x,balloon[k][detail-2].y,balloon[k][detail-2].z);
		}
	glNormal3f(normals[0][detail-2].x,normals[0][detail-2].y,normals[0][detail-2].z);
	glVertex3f(balloon[0][detail-2].x,balloon[0][detail-2].y,balloon[0][detail-2].z);
	glEnd();

	free(balloon);
}

