#include <math.h>			// Header File For Math Library Routines
#include <stdio.h>			// Header File For Standard I/O Routines
#include <stdlib.h>			// Header File For Standard Library
#include <iostream.h>

#include <gl\glut.h>
#include <windows.h>

#include "vector.h"
#include "tmatrix.h"

#include "const.h"
#include "noise.h"

struct Camera
{
	int xres, yres;
	Vector pos, tar;
	float rotx, roty, rotz;
	float tarx, tary, tarz;
};

Camera cam;
bool auto_rotate=false, normal=false, timer=false, fullscreen=true;

#define MIN(x,y) (x < y ? x : y)
#define MAX(x,y) (y > x ? y : x)

TMatrix rotate, rotate_cam = rotateY(.1), reverse_cam = rotateY(-.1);

#include "string.h"
#include "balloon.h"
#include "sky.h"

void drawScene ();

Balloon *first, *last;

int t=0;
float ang = 0;
float xang = 0;
int list_index[4];
int sky_list;
/////////////////////////////////////////////////////////////////////////////////////
//Everything else


void update_pos() //Defines the initial position of the camera
{
	cam.pos.x = 0;
	cam.pos.y = SIZE*3;
	cam.pos.z = -SIZE*2.5;
	
	cam.rotx = 0;
	cam.roty = 0;
	cam.rotz = 0;

	cam.tar.x = 0;
	cam.tar.y = SIZE*2.5;
	cam.tar.z = 0;
}

void init_lighting() //Various OGL calls to initialize lighting
{
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


	float light_ambient[] = {.5f, .5f, .5f, 1.0f};
	float light_diffuse[] = {1.0f, 1.0f, 1.0f, 1.0f};
	float light_specular[] = {1.0f, 1.0f, 1.0f, 1.0f};
	float light_position[] = {1.0f, 1.0f, 1.0f, 0.0f};
	float balloon_specular[] = {1.0f, 1.0f, 1.0f, 1.0f};
	float shiny_pants[] = {16.0f};

	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);

	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shiny_pants);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, balloon_specular);
}

void new_balloon(Balloon &newbie) //Adds a new balloon to the top of the linked list
{
	newbie.pos = Vector(((float)rand()/RAND_MAX)*DISPLACEMENT,0,((float)rand()/RAND_MAX)*DISPLACEMENT);
	newbie.string = String(NUM_KNOTS, newbie.pos);
	newbie.next = NULL;
}

void init_balloons() //Creates a complete list of balloons for the beginning of the program
{
	
	first = new Balloon;
	Balloon *curr = first;
	for (int x = 0; x < BALLOONS-1; x++)
	{
		curr->pos = Vector(((float)rand()/RAND_MAX)*DISPLACEMENT,(x*100),((float)rand()/RAND_MAX)*DISPLACEMENT);
		curr->string = String(NUM_KNOTS, curr->pos);
		curr->next = new Balloon;
		curr = curr->next;
	}
	curr->string = String(NUM_KNOTS, curr->pos);
	last = curr;
}



//the init function
void init(void)
{
	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glEnable(GL_DEPTH_TEST);
 
	glClearColor(0.2f, .5f, 1.0f, 1.0f);
	glClearDepth(1.0f);									
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
	glMatrixMode(GL_PROJECTION);
	gluPerspective(45.0f,(GLfloat)1024/(GLfloat)768,0.1f,1000000.0f);

	init_noise();

	glGenTextures(1, &texture);
	gen_clouds();

	init_balloons();
	
	for (int x = 1; x <= 4; x++) //Initialize balloon display lists
	{
		list_index[x-1] = glGenLists(x);
		if (list_index[x-1] != 0)
		{
			glNewList(list_index[x-1], GL_COMPILE);
				draw_balloon((35/(x*x))+8);
			glEndList();
		}
	}
	
	update_pos();
	init_lighting();
}


// the reshape function
void reshape(int w, int h) 
{
	if (h == 0)										// Prevent A Divide By Zero By
	{
		h = 1;										// Making Height Equal One
	}

	glViewport(0,0,w,h);								// Reset The Current Viewport

	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix

	// Calculate The Aspect Ratio Of The Window
	gluPerspective(45.0f,(GLfloat)w/(GLfloat)h,0.1f,1000000.0f);

	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix

	glutPostRedisplay() ;

	cam.xres = w;
	cam.yres = h;
}

//the display function
//This uses a (very) simple timer that keeps the animation going at a fairly constant speed
void display(void)
{
	unsigned long time1, time2;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT) ;
	
	if (timer)
	{
		time1 = timeGetTime();
		drawScene() ;
		time2 = timeGetTime();
		
		vel = (vel + MIN(.015*(time2-time1), .4))/2; //Limit vel to .4 so as to retain stability
		rotate_cam = rotateY(vel);
		reverse_cam = rotateY(-vel);
	}
	else 
	{
		drawScene();
	}		

	if (auto_rotate) cam.pos = rotate_cam * cam.pos;
	t++; //t is the 4th dimension in the turbulence function, and thus must be always increasing

	glutSwapBuffers() ;
}

void drawScene ()
{
	Vector temp;
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();		 						

	int dist, x;
	float alpha; 

	gluLookAt(cam.pos.x, cam.pos.y, cam.pos.z, cam.tar.x, cam.tar.y, cam.tar.z, 0.0f, 1.0f, 0.0f);
	
	if (first->pos.y > CEILING) //If a balloon has reached the ceiling, get rid of it
	{
		Balloon *temp;
		temp = first;
		first = first->next;
		delete temp;
	}

	//Update all the balloons
	for (Balloon *curr = first; curr != NULL; curr = curr->next)
	{
		curr->vel.x += ((turbulence(curr->pos.x*OFFSET, curr->pos.y*OFFSET, curr->pos.z*OFFSET, t, true)-.5)
			-curr->vel.x/inertia)*vel;
		curr->vel.y = BUOYANCY;
		curr->vel.z += ((turbulence(curr->pos.x*OFFSET, curr->pos.y*OFFSET, curr->pos.z*OFFSET, t, false)-.5)
			-curr->vel.z/inertia)*vel;
		
		curr->vel.x = MIN(curr->vel.x, 1);
		curr->vel.x = MAX(curr->vel.x, -1);
		curr->vel.z = MIN(curr->vel.z, 1);
		curr->vel.z = MAX(curr->vel.z, -1);
		
		curr->pos += curr->vel; 
		
		for (x = 0; x < MAX(vel/.07, 1); x++) //Yet another crude constant speed hack
			curr->string.update(curr->pos);

		if (curr->pos.y > HIGH_FADE) alpha = ((curr->pos.y - HIGH_FADE)/10);
		else if (curr->pos.y < LOW_FADE) alpha = ((LOW_FADE - curr->pos.y)/10);
		else alpha = 1;
		if (alpha != 0) alpha = 1/alpha;
		balloon_material[3] = alpha;

		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, balloon_material);


		glPushMatrix();
			glRotatef(ang,0,1,0);
			glRotatef(xang,1,0,0);
			glDisable(GL_LIGHTING);
				glColor4f(.7,.7,.7,alpha);
				curr->string.render();
			glEnable(GL_LIGHTING);
			glTranslatef(curr->pos.x, curr->pos.y, curr->pos.z);
			glRotatef(curr->vel.x*10,0,0,1);
			glRotatef(curr->vel.z*10,1,0,0);
			glScalef(30,30,30);

			dist = dist_squared(cam.pos, curr->pos);
			dist /= 100000;
			if (dist < 5 ) glCallList(list_index[0]);
			else if (dist < 15 ) glCallList(list_index[1]);
			else if (dist < 25 ) glCallList(list_index[2]);
			else glCallList(list_index[3]);
		glPopMatrix();

		if (curr == last && curr->pos.y > HEIGHT) //If the lowest balloon has reached a set height, create
		{										  //a new balloon
			curr->next = new Balloon;
			new_balloon(*curr->next);
			last = curr->next;
		}

	}

	//Draw the sky
	glPushMatrix(); 
		glRotatef(ang,0,1,0);
		glRotatef(xang,1,0,0);
		glTranslatef(0,-450000,0);
		render_sky(9); //A detail level of 9 is more than enough given the angle that the dome covers
	glPopMatrix();
}



void input_norm(unsigned char key, int x, int y)
{
	TMatrix trans;
	Vector dir;
	
	switch(key)
	{
	case 't':
		timer = !timer;
		vel = .07;
		break;
	case 'a':
		auto_rotate = !auto_rotate;
		break;
	case 'c':
		init_noise();
		gen_clouds();
		break;
	case '+':
		cam.pos = cam.tar + ((cam.pos-cam.tar)*.99);
		break;
	case '-':
		cam.pos = cam.tar + ((cam.pos-cam.tar)*1.01);
		break;
	case char(27):
		exit(1);
		break;
	}	
		
}

void input_special(int key, int x, int y)
{
	switch (key)
	{
	case GLUT_KEY_F1:
		fullscreen = !fullscreen;
		if (!fullscreen) 
		{
			glutReshapeWindow(1024, 768);//cam.yres);
			glutPositionWindow(10,30);
		}
		else glutFullScreen();
	case GLUT_KEY_UP:
		cam.tar.y += 10;
		break;
	case GLUT_KEY_DOWN:
		cam.tar.y -= 10;
		break;
	case GLUT_KEY_LEFT:
		cam.pos = reverse_cam * cam.pos;
		cam.pos = reverse_cam * cam.pos;
		break;
	case GLUT_KEY_RIGHT:
		cam.pos = rotate_cam * cam.pos;
		cam.pos = rotate_cam * cam.pos;
		break;
	}
}


		


int main(int argc, char** argv) //finally the main function
{
	//initialize glut
	glutInit(&argc, argv) ;
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH) ;   
	glutInitWindowSize(1024,768) ;
	
	GLint GLwin = glutCreateWindow("Red Balloon") ;
	glutFullScreen();
		
	glutDisplayFunc(display) ;
	glutReshapeFunc(reshape) ;
	glutKeyboardFunc(input_norm);
	glutSpecialFunc(input_special);
	glutIdleFunc(display) ;

	init() ;
	display();
		
	glutMainLoop() ;

	return 0;
}

